note
	description: "Huffman encoder."

class
	ENCODER

feature -- Main operations

	encode (text: STRING)
			-- Encode `text' and store result in `encoding'.
		require
			text_non_empty: text /= Void and then not text.is_empty
		do
			compute_probabilities (text)
			build_tree
			build_codes

			encoding := ""
			across
				text as c
			loop
				encoding.append (codes [c.item])
			end
		ensure
			encoding_non_empty: encoding /= Void and not encoding.is_empty
			encoding_binary: across encoding as c all c.item = '0' or c.item = '1' end
			code_tree_exists: code_tree /= Void
		end

feature -- Access

	encoding: STRING
			-- Result of the last encoding:
			-- a string consisting of charachers '0' and '1'
			-- (which could later be converted into a bitstring).

	code_tree: NODE
			-- Code tree built for the last encoding.

feature {NONE} -- Implementation

	probabilities: V_TABLE [CHARACTER, REAL_64]
			-- For each character in the text, the probability of its appearance.

	codes: V_TABLE [CHARACTER, STRING]
			-- For each character in the text, its Huffman code.

	compute_probabilities (text: STRING)
			-- Compute probablities for each character in `text'.
		require
			text_non_empty: text /= Void and then not text.is_empty
		do
			create {V_HASH_TABLE [CHARACTER, REAL_64]} probabilities
			across
				text as c
			loop
				if probabilities.has_key (c.item) then
					probabilities [c.item] := probabilities [c.item] + (1.0 / text.count)
				else
					probabilities [c.item] := (1.0 / text.count)
				end
			end
		ensure
			probablities_nonempty: probabilities /= Void and then not probabilities.is_empty
		end

	build_tree
			-- Build `code_tree' from `probablities'
		require
			probablities_nonempty: probabilities /= Void and then not probabilities.is_empty
		local
			nodes: V_DOUBLY_LINKED_LIST [NODE]
			nc: V_LIST_ITERATOR [NODE]
			l, r: NODE
		do
			-- Init nodes
			create nodes
			across
				probabilities as i
			loop
				nodes.extend_back (create {LEAF}.make (i.key, i.item))
			end
			-- Merge nodes
			from
				nc := nodes.new_cursor
			until
				nodes.count = 1
			loop
				find_min (nodes, nc)
				r := nc.item
				nc.remove
				find_min (nodes, nc)
				l := nc.item
				nc.remove
				nodes.extend_back (create {INTERNAL_NODE}.make (l, r))
			end
			code_tree := nodes.first
		ensure
			code_tree_exists: code_tree /= Void
		end

	find_min (l: V_LIST [NODE]; lc: V_ITERATOR [NODE])
			-- Move `lc' to the minimum element of `l'.
		require
			not_empty: l /= Void and then not l.is_empty
			lc_over_l: lc /= Void and then lc.target = l
		do
			lc.start
			across
				l as c
			loop
				if c.item.probability < lc.item.probability then
					lc.go_to (c.index)
				end
			end
		ensure
			lc_valid: not lc.off
		end

	build_codes
			-- Build `codes' from `code_tree'.
		require
			code_tree_exists: code_tree /= Void
		do
			create {V_HASH_TABLE [CHARACTER, STRING]} codes
			code_tree.fill_codes (codes, "")
		ensure
			codes_exists: codes /= Void
		end

feature {NONE} -- Output

	show_probablities
			-- Pretty-print `probabilities'.
		require
			probabilities_exists: probabilities /= Void
		do
			across
				probabilities as cur
			loop
				print (cur.key.out + " " + (cur.item * 100).rounded.out + "%%")
				io.new_line
			end
		end

	show_code_tree
			-- Pretty-print `code_tree'.
		require
			code_tree_exists: code_tree /= Void
		do
			code_tree.show ("")
		end

	show_codes
			-- Pretty-print `codes'.	
		require
			codes: codes /= Void
		do
			across
				codes as cur
			loop
				print (cur.key.out + " " + cur.item)
				io.new_line
			end
		end

end
